"""

    cacher is a Python module that enables easier integration with caching layer
    such as Memcached. It gives developers access to API that enables:

    1. decoration of python function which automates caching and invalidation.
    2. Batching API

"""

__version__ = '0.0.2'

from .cacher import Cacher

__all__ = ['Cacher']



class Batcher(object):
    """
    Batcher enables developers to batch multiple retrieval requests.

    Example usage #1::


        from pycacher import Cacher

        cacher = Cacher()
        batcher = cacher.create_batcher()

        batcher.add('testkey')
        batcher.add('testkey1')
        batcher.add('testkey2')

        values = batcher.batch()


    It is possible to use batcher as context manager. Inside the context manager,
    developers can call the `.register` method of cached functions
    to register its keys to the currently active batcher for later batching. Then,
    when the actual cached functions that were registered earlier inside the
    context manager were actually called, it will seek its value from the batcher
    context.

    
    Example usage #2::
        
        from pycacher import Cacher
        cacher = Cacher()

        batcher = cacher.create_batcher()
        
        with batcher:
            cached_func.register(1, 2)
            cached_func_2.register(1, 2)
            cached_func_3.register(3, 5)

        batcher.batch()

        #Later..

        with batcher:
            cached_func(1, 2) #will look for its value from the batcher
            cached_func_2(1, 2)
            cached_func(3, 5)

        #You can also do this:

        batcher.register(cached_func, 1, 2)
        batcher.register(cached_func_2, 1, 2)

    """
    
    def __init__(self, cacher=None):
        self.cacher = cacher
        self._keys = set()
        self._last_batched_values = {}

        self._autobatch_flag = False

    def add(self, key):

        if isinstance(key, list):
            for k in key:
                self._keys.add(k)
        else:
            self._keys.add(key)

    def reset(self):
        self._keys = set()

    def batch(self):
        self._last_batched_values = self.cacher.backend.multi_get(self._keys)

        return self._last_batched_values

    def register(self, decorated_func, *args):
        cache_key = decorated_func.build_cache_key(*args)

        self.add(cache_key)

    def get_last_batched_values(self):
        return self._last_batched_values

    def get_values(self):
        return self.get_last_batched_values()

    def get_keys(self):
        return self._keys

    def get(self, key):
        return self._last_batched_values.get(key)

    def is_batched(self, key):
        """Checks whether a key is included in the latest batch.
        
            Example:

            self.batcher.add('test-1')

            self.batcher.batch()

            self.batcher.is_batched('test-1')
            >> True

        """
        return key in self._last_batched_values

    def autobatch(self):
        """autobatch enables the batcher to automatically batch the batcher keys
        in the end of the context manager call.
            
        Example Usage::

                with batcher.autobatch():
                    expensive_function.register(1, 2)

                #is similar to:

                with batcher:
                    expensive_function.register(1, 2)

                batcher.batch()
                
        """

        self._autobatch_flag = True
        return self

    def __enter__(self):
        """ On context manager enter step, we're basically pushing this Batcher instance
        to the parent cacher's batcher context, so when the there are decorated
        functions that are registering, they know to which batcher to register to."""
        self.cacher.push_batcher(self)

    def __exit__(self, type, value, traceback):
        """ On exit, pop the batcher. """

        if self._autobatch_flag:
            self.batch()
            self._autobatch_flag = False
        self.cacher.pop_batcher()

class OutOfBatcherContextRegistrationException(Exception):
    pass



from functools import wraps
import pickle

from .backends import LocalBackend, MemcacheBackend
from .utils import default_cache_key_func
from .batcher import Batcher, OutOfBatcherContextRegistrationException

class Cacher(object):

    """

    Cacher is the the main object of the whole pycache packages, 
    which encapsulates connection with the Memcached client.

    Example instantiation::
    
        import pycacher

        cacher = pycacher.Cacher('localhost', 11211)
   

    By default, Cacher would be instantiated with `MemcacheBackend`. It's possible
    to instantiate Cacher manually with a different backend, such as LocalBackend.
    Here's how you do it::
        
        import pycacher
        from pycacher.backends import LocalBackend

        cacher = pycacher.Cacher(backend=LocalBackend())

    """
    def __init__(self, host='localhost', port=11211, client=None,
                       backend=None, default_expires=None, 
                       cache_key_func=default_cache_key_func):
        
        self.cache_key_func = cache_key_func 

        if backend:
            self.backend = backend
        else:
            self.backend = MemcacheBackend(host=host, port=port)
        
        self._batcher_ctx_stack = []

    def cache(self, expires=None):
        """Decorates a function to be cacheable.

        Example usage::
        
            @cacher.cache(expires=None)
            def expensive_function(a, b):
                pass

        """
        
        def decorator(f):

            #Wraps the function within a function decorator
            return CachedFunctionDecorator(f, cacher=self, expires=expires, 
                                              cache_key_func=self.cache_key_func)

        return decorator

    def create_batcher(self):
        """Simply creates a Batcher instance."""
        return Batcher(self)
    
    def push_batcher(self, batcher):
        self._batcher_ctx_stack.append(batcher)

    def get_current_batcher(self):

        if len(self._batcher_ctx_stack) > 0:
            return self._batcher_ctx_stack[-1]
        else:
            return None

    def pop_batcher(self):
        return self._batcher_ctx_stack.pop()

class CachedFunctionDecorator(object):
    
    def __init__(self, func, cacher=None, expires=None, 
                        cache_key_func=default_cache_key_func):
        self.func = func
        self.cacher = cacher
        self.cache_key_func = cache_key_func
        self.expires = expires

    def __call__(self, *args, **kwargs):
        """The method that will actually be called when the decorated functon
        is called."""

        cache_key = self._build_cache_key(*args)
        
        batcher = self.cacher.get_current_batcher()

        if batcher:
            unpickled_value = batcher.get(cache_key) or self.cacher.backend.get(cache_key)
        else:
            unpickled_value = self.cacher.backend.get(cache_key)

        if unpickled_value:
            return pickle.loads(unpickled_value)
        else:
            value = self.func(*args)
            self.cacher.backend.set(cache_key, pickle.dumps(value))
            return value

    def _build_cache_key(self, *args):
        """Builds the cache key with the supplied cache_key function """
        return self.cache_key_func(self.func, *args)

    def build_cache_key(self, *args):
        """Builds the cache key with the supplied cache_key function """
        return self.cache_key_func(self.func, *args)

    def warm(self, *args):
        """

            Forces to run the actual function (regardless of whether we already
            have the result on the cache or not) and set the backend to store
            the return value.

        """
        cache_key = self._build_cache_key(*args)

        value = self.func(*args)
        return self.cacher.backend.set(cache_key, pickle.dumps(value))

    def is_cached(self, *args):
        """
            Simply checks if the current function value with the supplied args
            is currently cached in the backend.
        """
        cache_key = self._build_cache_key(*args)

        return self.cacher.backend.exists(cache_key)

    def invalidate(self, *args):
        """Invalidates the current function's cache key with the current args.

        Example usage::

            is_user_board_subscriber.invalidate(uid, bid) 

        """
        return self.cacher.backend.delete(self._build_cache_key(*args))

    def register(self, *args):
        """Registers the cached function on an active batcher context for later batching.
            
            Example usage::

                batcher = cacher.create_batcher()

                with batcher:
                    cached_function.register(1, 2)

                batcher.batch()

                with batcher:
                    cached_function(1, 2) #now will look for its value in the current batcher's
                                          #last batched values.
        """
        batcher = self.cacher.get_current_batcher()

        if batcher:
            #Register the function and the args to the batcher 
            batcher.register(self, *args)
        else:
            raise OutOfBatcherContextRegistrationException()



import os
from distutils.core import setup

from pycacher import __version__

try:
    import setuptools
except ImportError:
    pass


with open(os.path.abspath('README.md')) as f:
    long_description = f.read()

setup(

    name="pycacher",
    version=__version__,
    packages=['pycacher'],
    author="Garindra Prahandono",
    install_requires=['python-memcached'],
    author_email="garindraprahandono@gmail.com",
    url='http://pycacher.readthedocs.org',
    download_url=('http://cloud.github.com/downloads/garindra/'
                      'pycacher/pycacher-%s.tar.gz' % __version__),
    description="pycacher is a python module which enables easy caching layer via function decorators, batcher, etc.",
    test_suite='nose.collector',
    long_description=long_description,
    classifiers=[
        'Development Status :: 3 - Alpha',
        'Environment :: Console',
        'Intended Audience :: Developers',
        'License :: OSI Approved :: MIT License',
        'Operating System :: OS Independent',
        'Programming Language :: Python']
)



"""
    
    This module contains different wrapper classes for different caching backends.

"""

try:
    import memcache
except ImportError:
    pass 

class Backend(object):
    
    def get(self, key):
        raise NotImplementedError("Backend subclasses should always implement `get` method")

    def set(self, key, value):
        raise NotImplementedError("Backend subclasses should always implement `set` method")

class LocalBackend(object):

    def __init__(self):
        
        self._dict = {}

    def set(self, key, value):
        self._dict[key] = value

    def get(self, key):
        return self._dict.get(key)

    def delete(self, key):
        del self._dict[key]

    def exists(self, key):
        return self._dict.get(key) is not None

    def multi_get(self, keys):
        
        values = {}

        for key in keys:
            values[key] = self._dict.get(key)

        return values

class MemcacheBackend(object):
    
    def __init__(self, client=None, host='127.0.0.1', port=11211):
        
        if client:
            self.client = client 
        else:
            self.client = memcache.Client([host+':'+str(port)], debug=0)

    def get(self, key):
        return self.client.get(key)

    def set(self, key, value):
        return self.client.set(key, value)

    def delete(self, key):
        return self.client.delete(key)

    def exists(self, key):
        return self.client.get(key) is not None

    def multi_get(self, keys):
        return self.client.get_multi(keys)

class PycacherBackendArgumentException(Exception):
    pass
